/*
 * Copyright (c) 2011 Picochip Ltd., Jamie Iles
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * All enquiries to support@picochip.com
 */
#ifndef __PICOXCELL_COMMON_H__
#define __PICOXCELL_COMMON_H__

#include <asm/mach/time.h>

extern struct sys_timer dw_apb_timer;

#endif /* __PICOXCELL_COMMON_H__ */
